//
//  Mail.swift
//  Hedwig
//
//  Created by Wei Wang on 2016/12/31.
//
//  Copyright (c) 2017 Wei Wang <onev@onevcat.com>
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.

import Foundation
import AddressParser

public enum MailError: Error {
    case noSender
    case noRecipient
}

public struct Mail {
    public let from: NameAddressPair?
    public let to: [NameAddressPair]
    public let cc: [NameAddressPair]?
    public let bcc: [NameAddressPair]?
    public let subject: String
    public let text: String
    public let attachments: [Attachment]
    public let additionalHeaders: [String: String]

    public let alternative: Attachment?
    
    public let messageId = UUID().uuidString
    public let date = Date()
    
    public init(text: String,
         from: String,
         to: String,
         cc: String? = nil,
         bcc: String? = nil,
         subject: String = "",
         attachments: [Attachment]? = nil,
         additionalHeaders: [String: String] = [:])
    {
        self.from = from.parsedAddresses.last

        self.to = to.parsedAddresses
        self.cc = cc?.parsedAddresses
        self.bcc = bcc?.parsedAddresses
        
        self.text = text
        self.subject = subject

        if let attachments = attachments {
            let result = attachments.takeLast { $0.isAlternative }
            self.alternative = result.0
            self.attachments = result.1
        } else {
            self.alternative = nil
            self.attachments = []
        }
    
        self.additionalHeaders = additionalHeaders
    }
    
    var hasSender: Bool {
        return from != nil
    }
    
    var hasRecipient: Bool {
        let noCc = self.cc?.isEmpty ?? true
        let noBcc = self.bcc?.isEmpty ?? true
        let noRecipient = to.isEmpty && noCc && noBcc
        return !noRecipient
    }
}

extension Mail {
    private var headers: [String: String] {
        var fields = [String: String]()
        fields["MESSAGE-ID"] = messageId
        fields["DATE"] = date.smtpFormatted
        fields["FROM"] = from?.mime ?? ""
    
        fields["TO"] = to.map { $0.mime }.joined(separator: ", ")
        
        if let cc = cc {
            fields["CC"] = cc.map { $0.mime }.joined(separator: ", ")
        }
        
        fields["SUBJECT"] = subject.mimeEncoded ?? ""
        fields["MIME-VERSION"] = "1.0 (Hedwig)"
        
        for (key, value) in additionalHeaders {
            fields[key.uppercased()] = value
        }
    
        return fields
    }
    
    var headersString: String {
        return headers.map { (key, value) in
            return "\(key): \(value)"
        }.joined(separator: CRLF)
    }
    
    var hasAttachment: Bool {
        return alternative != nil || !attachments.isEmpty
    }
}

public struct NameAddressPair {
    public let name: String
    public let address: String
    
    public init(name: String, address: String) {
        self.name = name
        self.address = address
    }
    
    var mime: String {
        if name.isEmpty {
            return address
        }
        
        if let nameEncoded = name.mimeEncoded {
            return "\(nameEncoded) <\(address)>"
        } else {
            return address
        }
    }
}

extension Address {
    var allMails: [NameAddressPair] {
        var result = [NameAddressPair]()
        switch entry {
        case .mail(let address):
            result.append(NameAddressPair(name: name, address: address))
        case .group(let addresses):
            let converted = addresses.flatMap { return $0.allMails }
            result.append(contentsOf: converted)
        }
        return result
    }
}

extension String {
    var parsedAddresses: [NameAddressPair] {
        return AddressParser.parse(self).flatMap { $0.allMails }
    }
}

extension String {
    // A simple but maybe not fully compatible with RFC 2047
    // https://tools.ietf.org/html/rfc2047
    var mimeEncoded: String? {
        guard let encoded = addingPercentEncoding(
            withAllowedCharacters: .urlQueryAllowed) else {
            return nil
        }
        
        let quoted = encoded
            .replacingOccurrences(of: "%20", with: "_")
            .replacingOccurrences(of: ",", with: "%2C")
            .replacingOccurrences(of: "%", with: "=")
        return "=?UTF-8?Q?\(quoted)?="
    }
}

extension DateFormatter {
    static let smtpDateFormatter: DateFormatter = {
        let formatter = DateFormatter()
        formatter.locale = Locale(identifier: "en-US")
        formatter.dateFormat = "EEE, d MMM yyyy HH:mm:ss ZZZ"
        return formatter
    }()
}

extension Date {
    var smtpFormatted: String {
        return DateFormatter.smtpDateFormatter.string(from: self)
    }
}

extension Array {
    func takeLast(where condition: (Element) -> Bool) -> (Element?, Array) {
        var index: Int? = nil
        for i in (0 ..< count).reversed() {
            if condition(self[i]) {
                index = i
                break
            }
        }
        
        if let index = index {
            var array = self
            let ele = array.remove(at: index)
            return (ele, array)
        } else {
            return (nil, self)
        }
    }
}
